@@ -25,12 +25,15 @@
 #include <fcntl.h>
 #include "irc.h"
 #include "conf.h"
+#include "path_util.h"
 #include "tuple.h"
 #include "log.h"
 #include "bip.h"
 #include "line.h"
 #include "defaults.h"
 
+#define S_CONF "bip.conf"
+
 int sighup = 0;
 
 char *conf_log_root;
@@ -41,16 +44,23 @@ unsigned short conf_port;
 int conf_css;
 #ifdef HAVE_LIBSSL
 char *conf_ssl_certfile;
+char *conf_client_ciphers;
+char *conf_client_dh_file;
+char *conf_server_default_ciphers;
 #endif
 int conf_daemonize;
 char *conf_pid_file;
 char *conf_biphome;
+int conf_reconn_timer;
 
 /* log options, for sure the trickiest :) */
 int conf_log = DEFAULT_LOG;
 int conf_log_system = DEFAULT_LOG_SYSTEM;
 int conf_log_sync_interval = DEFAULT_LOG_SYNC_INTERVAL;
 
+bip_t *_bip;
+FILE *conf_global_log_file;
+
 list_t *parse_conf(FILE *file, int *err);
 void conf_die(bip_t *bip, char *fmt, ...);
 static char *get_tuple_pvalue(list_t *tuple_l, int lex);
@@ -58,9 +68,6 @@ void bip_notify(struct link_client *ic, char *fmt, ...);
 void adm_list_connections(struct link_client *ic, struct bipuser *bu);
 void free_conf(list_t *l);
 
-#ifdef HAVE_OIDENTD
-#define OIDENTD_FILENAME ".oidentd.conf"
-#endif
 
 static void hash_binary(char *hex, unsigned char **password, unsigned int *seed)
 {
@@ -87,6 +94,7 @@ static void hash_binary(char *hex, unsigned char **password, unsigned int *seed)
 	sscanf(hex + 2 * 3, "%02x", &buf);
 	*seed |= buf;
 
+	MAYFREE(*password);
 	*password = md5;
 }
 
@@ -150,58 +158,28 @@ void conf_die(bip_t *bip, char *fmt, ...)
 	va_end(ap);
 }
 
-FILE *conf_global_log_file;
-
-static pid_t daemonize(void)
-{
-	switch (fork()) {
-	case -1:
-		fatal("Fork failed");
-		break;
-	case 0:
-		break;
-	default:
-		_exit(0);
-	}
-
-	if (setsid() < 0)
-		fatal("setsid() failed");
-
-	switch (fork()) {
-	case -1:
-		fatal("Fork failed");
-		break;
-	case 0:
-		break;
-	default:
-		_exit(0);
-	}
-
-	close(0);
-	close(1);
-	close(2);
-	/* This better be the very last action since fatal makes use of
-	 * conf_global_log_file */
-	return getpid();
-}
 
 /* RACE CONDITION! */
 int do_pid_stuff(void)
 {
-	char hname[1024];
-	char longpath[1024];
+	char hname[512];
 	FILE *f;
 	int fd;
+	// size is conf_pid_file + hname max + %ld max + two '.'.
+	int longpath_max = strlen(conf_pid_file) + 512 + 3 + 20;
+	char *longpath = bip_malloc(longpath_max + 1);
 
 try_again:
 	fd = -1;
 	f = fopen(conf_pid_file, "r");
 	if (f)
 		goto pid_is_there;
-	if (gethostname(hname, 1023) == -1)
+	if (gethostname(hname, 511) == -1)
 		fatal("%s %s", "gethostname", strerror(errno));
-	snprintf(longpath, 1023, "%s.%s.%ld", conf_pid_file, hname,
+	hname[511] = 0;
+	snprintf(longpath, longpath_max - 1, "%s.%s.%ld", conf_pid_file, hname,
 			(long unsigned int)getpid());
+	longpath[longpath_max] = 0;
 	if ((fd = open(longpath, O_CREAT|O_WRONLY, S_IWUSR|S_IRUSR)) == -1)
 		fatal("Cannot write to PID file (%s) %s", longpath,
 				strerror(errno));
@@ -215,6 +193,7 @@ try_again:
 		}
 	}
 	unlink(longpath);
+	free(longpath);
 	return fd;
 pid_is_there:
 	{
@@ -256,33 +235,10 @@ pid_is_there:
 					" %s.", conf_pid_file);
 		exit(2);
 	}
+	free(longpath);
 	return 0;
 }
 
-#define S_CONF "bip.conf"
-
-static void usage(char *name)
-{
-	printf(
-"Usage: %s [-f config_file] [-h] [-n]\n"
-"	-f config_file: Use config_file as the configuration file\n"
-"		If no config file is given %s will try to open ~/.bip/" S_CONF "\n"
-"	-n: Don't daemonize, log in stderr\n"
-"	-v: Print version and exit\n"
-"	-h: This help\n", name, name);
-	exit(1);
-}
-
-static void version()
-{
-	printf(
-"Bip IRC Proxy - " PACKAGE_VERSION "\n"
-"Copyright © Arnaud Cornet and Loïc Gomez (2004 - 2008)\n"
-"Distributed under the GNU Public License Version 2\n");
-}
-
-bip_t *_bip;
-
 void reload_config(int i)
 {
 	(void)i;
@@ -355,6 +311,9 @@ static int add_network(bip_t *bip, list_t *data)
 		case LEX_SSL:
 			n->ssl = t->ndata;
 			break;
+		case LEX_CIPHERS:
+			MOVE_STRING(n->ciphers, t->pdata);
+			break;
 #endif
 		case LEX_SERVER:
 			n->serverv = bip_realloc(n->serverv, (n->serverc + 1)
@@ -380,6 +339,20 @@ static int add_network(bip_t *bip, list_t *data)
 			free(t->pdata);
 		free(t);
 	}
+
+#ifdef HAVE_LIBSSL
+	if (!n->ciphers) {
+		n->ciphers = conf_server_default_ciphers;
+	}
+#endif
+	if (n->serverc == 0) {
+		conf_die(bip, "No server in network: %s", n->name);
+		hash_remove_if_exists(&bip->networks, name);
+		free(n->name);
+		free(n);
+		n = NULL;
+		return 0;
+	}
 	return 1;
 }
 
@@ -652,12 +625,22 @@ static int get_tuple_nvalue(list_t *tuple_l, int lex)
 	return -1;
 }
 
+enum BLTimestamp lex_backlog_timestamp(char * tdata) {
+	if (strcmp(tdata, "time") == 0) {
+		return BLTSTime;
+	} else if (strcmp(tdata, "datetime") == 0) {
+		return BLTSDateTime;
+	} else {
+		return BLTSNone;
+	}
+}
+
 struct historical_directives {
 	int always_backlog;
 	int backlog;
 	int bl_msg_only;
 	int backlog_lines;
-	int backlog_no_timestamp;
+	enum BLTimestamp backlog_timestamp;
 	int blreset_on_talk;
 };
 
@@ -685,27 +668,17 @@ static int add_user(bip_t *bip, list_t *data, struct historical_directives *hds)
 		u->always_backlog = DEFAULT_ALWAYS_BACKLOG;
 		u->bl_msg_only = DEFAULT_BL_MSG_ONLY;
 		u->backlog_lines = DEFAULT_BACKLOG_LINES;
-		u->backlog_no_timestamp = DEFAULT_BACKLOG_NO_TIMESTAMP;
+		u->backlog_timestamp = DEFAULT_BACKLOG_TIMESTAMP;
 		u->blreset_on_talk = DEFAULT_BLRESET_ON_TALK;
 		u->blreset_connection = DEFAULT_BLRESET_CONNECTION;
 		u->bip_use_notice = DEFAULT_BIP_USE_NOTICE;
-	} else {
-		FREE(u->name);
-		FREE(u->password);
-		FREE(u->default_nick);
-		FREE(u->default_username);
-		FREE(u->default_realname);
-#ifdef HAVE_LIBSSL
-		FREE(u->ssl_check_store);
-		FREE(u->ssl_client_certfile);
-#endif
 	}
 
 	u->backlog = hds->backlog;
 	u->always_backlog = hds->always_backlog;
 	u->bl_msg_only = hds->bl_msg_only;
 	u->backlog_lines = hds->backlog_lines;
-	u->backlog_no_timestamp = hds->backlog_no_timestamp;
+	u->backlog_timestamp = hds->backlog_timestamp;
 	u->blreset_on_talk = hds->blreset_on_talk;
 
 	while ((t = list_remove_first(data))) {
@@ -743,7 +716,10 @@ static int add_user(bip_t *bip, list_t *data, struct historical_directives *hds)
 			u->backlog_lines = t->ndata;
 			break;
 		case LEX_BACKLOG_NO_TIMESTAMP:
-			u->backlog_no_timestamp = t->ndata;
+			u->backlog_timestamp = t->ndata ? BLTSNone : BLTSTime;
+			break;
+		case LEX_BACKLOG_TIMESTAMP:
+			u->backlog_timestamp = lex_backlog_timestamp(t->pdata);
 			break;
 		case LEX_BLRESET_ON_TALK:
 			u->blreset_on_talk = t->ndata;
@@ -865,11 +841,18 @@ static int validate_config(bip_t *bip)
 		}
 	}
 
-	hash_it_init(&bip->users, &it);
-	hash_it_next(&it);
-	if (hash_it_item(&it) && !strstr(conf_log_format, "%u"))
-		mylog(LOG_WARN, "log_format does not contain %%u, all users'"
-			" logs will be mixed !");
+	if (!strstr(conf_log_format, "%u")) {
+		hash_it_init(&bip->users, &it);
+		if (hash_it_item(&it)) {
+			// hash contains at least one element
+			hash_it_next(&it);
+			if (hash_it_item(&it)) {
+				// hash contains at least two elements
+				mylog(LOG_WARN, "log_format does not contain %%u, all users'"
+					" logs will be mixed !");
+			}
+		}
+	}
 	return r;
 }
 
@@ -959,7 +942,7 @@ int fireup(bip_t *bip, FILE *conf)
 	SET_HV(hds.backlog, BACKLOG);
 	SET_HV(hds.bl_msg_only, BL_MSG_ONLY);
 	SET_HV(hds.backlog_lines, BACKLOG_LINES);
-	SET_HV(hds.backlog_no_timestamp, BACKLOG_NO_TIMESTAMP);
+	SET_HV(hds.backlog_timestamp, BACKLOG_TIMESTAMP);
 	SET_HV(hds.blreset_on_talk, BLRESET_ON_TALK);
 #undef SET_HV
 
@@ -989,16 +972,31 @@ int fireup(bip_t *bip, FILE *conf)
 		case LEX_PORT:
 			conf_port = t->ndata;
 			break;
+		case LEX_RECONN_TIMER:
+			conf_reconn_timer = t->ndata;
+			break;
 #ifdef HAVE_LIBSSL
+		case LEX_DEFAULT_CIPHERS:
+			MOVE_STRING(conf_server_default_ciphers, t->pdata);
+			break;
+		case LEX_CSS_CIPHERS:
+			MOVE_STRING(conf_client_ciphers, t->pdata);
+			break;
 		case LEX_CSS:
 			conf_css = t->ndata;
 			break;
 		case LEX_CSS_PEM:
 			MOVE_STRING(conf_ssl_certfile, t->pdata);
 			break;
+		case LEX_DH_PARAM:
+			MOVE_STRING(conf_client_dh_file, t->pdata);
+			break;
 #else
+		case LEX_DEFAULT_CIPHERS:
 		case LEX_CSS:
+		case LEX_CSS_CIPHERS:
 		case LEX_CSS_PEM:
+		case LEX_DH_PARAM:
 			mylog(LOG_WARN, "Found SSL option whereas bip is "
 					"not built with SSL support.");
 			break;
@@ -1006,6 +1004,12 @@ int fireup(bip_t *bip, FILE *conf)
 		case LEX_PID_FILE:
 			MOVE_STRING(conf_pid_file, t->pdata);
 			break;
+		case LEX_WRITE_OIDENTD:
+			bip->write_oidentd = t->ndata;
+			break;
+		case LEX_OIDENTD_FILE:
+			MOVE_STRING(bip->oidentdpath, t->pdata);
+			break;
 		case LEX_ALWAYS_BACKLOG:
 			hds.always_backlog = t->ndata;
 			break;
@@ -1019,7 +1023,10 @@ int fireup(bip_t *bip, FILE *conf)
 			hds.backlog_lines = t->ndata;
 			break;
 		case LEX_BACKLOG_NO_TIMESTAMP:
-			hds.backlog_no_timestamp = t->ndata;
+			hds.backlog_timestamp = t->ndata ? BLTSNone : BLTSTime;
+			break;
+		case LEX_BACKLOG_TIMESTAMP:
+			hds.backlog_timestamp = lex_backlog_timestamp(t->pdata);
 			break;
 		case LEX_BLRESET_ON_TALK:
 			hds.blreset_on_talk = t->ndata;
@@ -1060,23 +1067,6 @@ out_conf_error:
 	return 0;
 }
 
-static void log_file_setup(void)
-{
-	char buf[4096];
-
-	if (conf_log_system && conf_daemonize) {
-		if (conf_global_log_file && conf_global_log_file != stderr)
-			fclose(conf_global_log_file);
-		snprintf(buf, 4095, "%s/bip.log", conf_log_root);
-		FILE *f = fopen(buf, "a");
-		if (!f)
-			fatal("Can't open %s: %s", buf, strerror(errno));
-		conf_global_log_file = f;
-	} else {
-		conf_global_log_file = stderr;
-	}
-}
-
 void check_rlimits()
 {
 	int r, cklim;
@@ -1152,185 +1142,6 @@ void check_rlimits()
 		mylog(LOG_WARN, "You can check your limits with `ulimit -a'");
 }
 
-int main(int argc, char **argv)
-{
-	FILE *conf = NULL;
-	char *confpath = NULL;
-	int ch;
-	int r, fd;
-	char buf[30];
-	bip_t bip;
-
-	bip_init(&bip);
-	_bip = &bip;
-
-	conf_ip = bip_strdup("0.0.0.0");
-	conf_port = 7778;
-	conf_css = 0;
-
-	signal(SIGPIPE, SIG_IGN);
-	signal(SIGHUP, reload_config);
-	signal(SIGINT, bad_quit);
-	signal(SIGQUIT, bad_quit);
-	signal(SIGTERM, bad_quit);
-	signal(SIGXFSZ, rlimit_bigfile_reached);
-	signal(SIGXCPU, rlimit_cpu_reached);
-
-	conf_log_root = NULL;
-	conf_log_format = bip_strdup(DEFAULT_LOG_FORMAT);
-	conf_log_level = DEFAULT_LOG_LEVEL;
-	conf_daemonize = 1;
-	conf_global_log_file = stderr;
-	conf_pid_file = NULL;
-#ifdef HAVE_LIBSSL
-	conf_ssl_certfile = NULL;
-#endif
-
-	while ((ch = getopt(argc, argv, "hvnf:s:")) != -1) {
-		switch (ch) {
-		case 'f':
-			confpath = bip_strdup(optarg);
-			break;
-		case 'n':
-			conf_daemonize = 0;
-			break;
-		case 's':
-			conf_biphome = bip_strdup(optarg);
-			break;
-		case 'v':
-			version();
-			exit(0);
-			break;
-		default:
-			version();
-			usage(argv[0]);
-		}
-	}
-
-	umask(0027);
-
-	check_rlimits();
-
-	char *home = NULL; /* oidentd path searching ignores conf_biphome */
-	home = getenv("HOME");
-	if (!home) {
-		conf_die(&bip, "no $HOME !, do you live in a trailer ?");
-		return 0;
-	}
-#ifdef HAVE_OIDENTD
-	bip.oidentdpath = bip_malloc(strlen(home) + 1 +
-			strlen(OIDENTD_FILENAME) + 1);
-	strcpy(bip.oidentdpath, home);
-	strcat(bip.oidentdpath, "/");
-	strcat(bip.oidentdpath, OIDENTD_FILENAME);
-#endif
-
-
-	if (!conf_biphome) {
-		conf_biphome = bip_malloc(strlen(home) + strlen("/.bip") + 1);
-		strcpy(conf_biphome, home);
-		strcat(conf_biphome, "/.bip");
-	}
-
-	if (!confpath) {
-		confpath = bip_malloc(strlen(conf_biphome) + 1 +
-				strlen(S_CONF) + 1);
-		strcpy(confpath, conf_biphome);
-		strcat(confpath, "/");
-		strcat(confpath, S_CONF);
-	}
-	conf = fopen(confpath, "r");
-	if (!conf)
-		fatal("config file not found");
-
-	r = fireup(&bip, conf);
-	fclose(conf);
-	if (!r)
-		fatal("Not starting: error in config file.");
-
-	if (!conf_log_root) {
-		char *ap = "/logs";
-		conf_log_root = bip_malloc(strlen(conf_biphome) +
-				strlen(ap) + 1);
-		strcpy(conf_log_root, conf_biphome);
-		strcat(conf_log_root, ap);
-		mylog(LOG_INFO, "Default log root: %s", conf_log_root);
-	}
-	if (!conf_pid_file) {
-		char *pid = "/bip.pid";
-		conf_pid_file = bip_malloc(strlen(conf_biphome) +
-				strlen(pid) + 1);
-		strcpy(conf_pid_file, conf_biphome);
-		strcat(conf_pid_file, pid);
-		mylog(LOG_INFO, "Default pid file: %s", conf_pid_file);
-	}
-
-#ifdef HAVE_LIBSSL
-	if (conf_css) {
-		int e, fd;
-		struct stat fs;
-
-		if (!conf_ssl_certfile) {
-			char *ap = "/bip.pem";
-			conf_ssl_certfile = bip_malloc(strlen(conf_biphome) +
-					strlen(ap) + 1);
-			strcpy(conf_ssl_certfile, conf_biphome);
-			strcat(conf_ssl_certfile, ap);
-			mylog(LOG_INFO, "Using default SSL certificate file: "
-					"%s", conf_ssl_certfile);
-		}
-
-		if ((fd = open(conf_ssl_certfile, O_RDONLY)) == -1)
-			fatal("Unable to open PEM file %s for reading",
-				conf_ssl_certfile);
-		else
-			close(fd);
-
-		e = stat(conf_ssl_certfile, &fs);
-		if (e)
-			mylog(LOG_WARN, "Unable to check PEM file, stat(%s): "
-				"%s", conf_ssl_certfile, strerror(errno));
-		else if ((fs.st_mode & S_IROTH) | (fs.st_mode & S_IWOTH))
-			mylog(LOG_ERROR, "PEM file %s should not be world "
-				"readable / writable. Please fix the modes.",
-				conf_ssl_certfile);
-	}
-#endif
-
-	check_dir(conf_log_root, 1);
-	fd = do_pid_stuff();
-	pid_t pid = 0;
-
-	log_file_setup();
-	if (conf_daemonize)
-		pid = daemonize();
-	else
-		pid = getpid();
-	snprintf(buf, 29, "%ld\n", (long unsigned int)pid);
-	write(fd, buf, strlen(buf));
-	close(fd);
-
-	bip.listener = listen_new(conf_ip, conf_port, conf_css);
-	if (!bip.listener || bip.listener->connected == CONN_ERROR)
-		fatal("Could not create listening socket");
-
-	for (;;) {
-		irc_main(&bip);
-
-		sighup = 0;
-
-		conf = fopen(confpath, "r");
-		if (!conf)
-			fatal("%s config file not found", confpath);
-		fireup(&bip, conf);
-		fclose(conf);
-
-		/* re-open to allow logfile rotate */
-		log_file_setup();
-	}
-	return 1;
-}
-
 #define RET_STR_LEN 256
 #define LINE_SIZE_LIM 70
 void adm_print_connection(struct link_client *ic, struct link *lnk,
@@ -1540,9 +1351,15 @@ noroom:
 	bip_notify(ic, "%s", buf);
 
 #ifdef HAVE_LIBSSL
-	bip_notify(ic, "SSL check mode '%s', stored into '%s'",
-		   checkmode2text(u->ssl_check_mode),
-		   STRORNULL(u->ssl_check_store));
+	if (u->ssl_check_store) {
+		bip_notify(ic, "SSL check mode '%s', stored into '%s'",
+				checkmode2text(u->ssl_check_mode),
+				u->ssl_check_store);
+	}
+	else {
+		bip_notify(ic, "SSL check mode '%s', default or no certificate store",
+				checkmode2text(u->ssl_check_mode));
+	}
 	if (u->ssl_client_certfile)
 		bip_notify(ic, "SSL client certificate stored into '%s'",
 				u->ssl_client_certfile);
@@ -1551,9 +1368,10 @@ noroom:
 		   STRORNULL(u->default_nick), STRORNULL(u->default_username),
 		   STRORNULL(u->default_realname));
 	if (u->backlog) {
-		bip_notify(ic, "Backlog enabled, lines: %d, no timestamp: %s,"
+		bip_notify(ic, "Backlog enabled, lines: %d, timestamp: %s,"
 			"  messages only: %s", u->backlog_lines,
-			bool2text(u->backlog_no_timestamp),
+			u->backlog_timestamp == BLTSNone ? "none" :
+				(u->backlog_timestamp == BLTSTime ? "time" : "datetime"),
 			bool2text(u->bl_msg_only));
 		bip_notify(ic, "always backlog: %s, reset on talk: %s",
 			bool2text(u->always_backlog),
@@ -1812,7 +1630,6 @@ int adm_trust(struct link_client *ic, struct line *line)
 				"connecting your client again.\r\n");
 
 		LINK(ic)->recon_timer = 1; /* Speed up reconnection... */
-		unbind_from_link(ic);
 		return OK_CLOSE;
 	}
 	return OK_FORGET;
@@ -2255,8 +2072,10 @@ int adm_bip(bip_t *bip, struct link_client *ic, struct line *line, int privmsg)
 			adm_bip_delconn(bip, ic,
 					irc_line_elem(line, privmsg + 2));
 		}
+#ifdef HAVE_LIBSSL
 	} else if (strcasecmp(irc_line_elem(line, privmsg + 1), "TRUST") == 0) {
 		return adm_trust(ic, line);
+#endif
 	} else {
 		bip_notify(ic, "Unknown command.");
 	}
