@@ -1,4 +1,3 @@
-
 /*
  * $Id: irc.c,v 1.156 2005/04/21 06:58:50 nohar Exp $
  *
@@ -31,6 +30,8 @@ extern bip_t *_bip;
 static int irc_join(struct link_server *server, struct line *line);
 static int irc_part(struct link_server *server, struct line *line);
 static int irc_mode(struct link_server *server, struct line *line);
+static int irc_mode_channel(struct link_server *s, struct channel *channel,
+				struct line *line, const char* mode, int add, int cur_arg);
 static int irc_kick(struct link_server *server, struct line *line);
 static int irc_privmsg(struct link_server *server, struct line *line);
 static int irc_notice(struct link_server *server, struct line *line);
@@ -47,14 +48,17 @@ static int irc_368(struct link_server *server, struct line *l);
 void irc_server_shutdown(struct link_server *s);
 static int origin_is_me(struct line *l, struct link_server *server);
 static void ls_set_nick(struct link_server *ircs, char *nick);
+static void server_set_chanmodes(struct link_server *l, const char *chanmodes);
+static void server_set_prefix(struct link_server *l, const char *prefix);
+static void server_init_modes(struct link_server *s);
+static int bip_get_index(const char* str, char car);
+static int bip_fls(int v);
 
-#ifdef HAVE_OIDENTD
-#define OIDENTD_FILENAME ".oidentd.conf"
 void oidentd_dump(bip_t *bip);
-#endif
 
 void irc_client_free(struct link_client *cli);
 extern int conf_log_sync_interval;
+extern int conf_reconn_timer;
 
 void write_user_list(connection_t *c, char *dest);
 
@@ -66,7 +70,6 @@ int irc_cli_bip(bip_t *bip, struct link_client *ic, struct line *line);
 
 #define LAGOUT_TIME 480
 #define LAGCHECK_TIME (90)
-#define RECONN_TIMER (120)
 #define RECONN_TIMER_MAX (600)
 #define LOGGING_TIMEOUT (360)
 #define CONN_INTERVAL 60
@@ -102,7 +105,7 @@ char *nick_from_ircmask(const char *mask)
 
 #define NAMESIZE 256
 
-list_t *channel_name_list(struct channel *c)
+list_t *channel_name_list(struct link_server *server, struct channel *c)
 {
 	list_t *ret;
 	hash_iterator_t hi;
@@ -128,13 +131,13 @@ list_t *channel_name_list(struct channel *c)
 			strcat(str, " ");
 			len++;
 		}
-		if (ovmask & NICKOP)
-			strcat(str, "@");
-		else if (ovmask & NICKHALFOP)
-			strcat(str, "%");
-		else if (ovmask & NICKVOICED)
-			strcat(str, "+");
-		len++;
+
+		// prepend symbol corresponding to the usermode
+		int msb;
+		if ((msb = bip_fls(ovmask))) {
+			str[len] = server->prefixes[msb - 1];
+			str[++len] = 0;
+		}
 
 		strcat(str, nick);
 		len += strlen(nick);
@@ -308,8 +311,9 @@ static int irc_352(struct link_server *server, struct line *line)
 	return OK_COPY_WHO;
 }
 
-static int irc_315(struct link_server *server, UNUSED(struct line *l))
+static int irc_315(struct link_server *server, struct line *l)
 {
+	(void)l;
 	struct link *link = LINK(server);
 	if (link->who_client) {
 		if (link->who_client->who_count == 0) {
@@ -417,12 +421,17 @@ int irc_dispatch_server(bip_t *bip, struct link_server *server,
 
 	} else if (LINK(server)->s_state == IRCS_CONNECTING) {
 		ret = OK_FORGET;
-		if (LINK(server)->ignore_server_capab &&
-				irc_line_elem_equals(line, 0, "005")) {
+			if (irc_line_elem_equals(line, 0, "005")) {
 			int i;
-			for (i = 0; i < irc_line_count(line); i++)
-				if (irc_line_elem_equals(line, i, "CAPAB"))
+			for (i = irc_line_count(line) - 1; i > 0; i--) {
+				if (LINK(server)->ignore_server_capab &&
+						irc_line_elem_equals(line, i, "CAPAB"))
 					irc_line_drop(line, i);
+				else if (!strncmp(irc_line_elem(line, i), "CHANMODES=", 10))
+					server_set_chanmodes(server, irc_line_elem(line, i) + 10);
+				else if (!strncmp(irc_line_elem(line, i), "PREFIX=(", 8))
+					server_set_prefix(server, irc_line_elem(line, i) + 7);
+			}
 		}
 		if (irc_line_elem_equals(line, 0, "NOTICE")) {
 		} else if (irc_line_elem_equals(line, 0, "376")) {
@@ -534,7 +543,7 @@ static void irc_send_join(struct link_client *ic, struct channel *chan)
 		WRITE_LINE4(CONN(ic), P_SERV, "333", LINK(ic)->l_server->nick,
 				chan->name, chan->creator, chan->create_ts);
 
-	list_t *name_list = channel_name_list(chan);
+	list_t *name_list = channel_name_list(LINK(ic)->l_server, chan);
 	char *s;
 	while ((s = list_remove_first(name_list))) {
 		char tmptype[2];
@@ -622,7 +631,7 @@ static char *get_str_elem(char *str, int num)
 		if (c - cur < 1)
 			return NULL;
 		ret = bip_malloc(c - cur + 1);
-		strncpy(ret, cur, c - cur);
+		memcpy(ret, cur, c - cur);
 		ret[c - cur] = 0;
 		return ret;
 	}
@@ -631,7 +640,7 @@ static char *get_str_elem(char *str, int num)
 		if (c - cur < 1)
 			return NULL;
 		ret = bip_malloc(c - cur + 1);
-		strncpy(ret, cur, c - cur);
+		memcpy(ret, cur, c - cur);
 		ret[c - cur] = 0;
 		return ret;
 	}
@@ -1226,7 +1235,7 @@ static void irc_copy_cli(struct link_client *src, struct link_client *dest,
 	return;
 }
 
-static int irc_dispatch_loging_client(bip_t *bip, struct link_client *ic,
+static int irc_dispatch_logging_client(bip_t *bip, struct link_client *ic,
 		struct line *line)
 {
 	if (irc_line_count(line) == 0)
@@ -1251,8 +1260,8 @@ int irc_dispatch(bip_t *bip, struct link_any *l, struct line *line)
 	case IRC_TYPE_CLIENT:
 		return irc_dispatch_client(bip, (struct link_client*)l, line);
 		break;
-	case IRC_TYPE_LOGING_CLIENT:
-		return irc_dispatch_loging_client(bip, (struct link_client*)l,
+	case IRC_TYPE_LOGGING_CLIENT:
+		return irc_dispatch_logging_client(bip, (struct link_client*)l,
 				line);
 		break;
 #ifdef HAVE_LIBSSL
@@ -1384,33 +1393,14 @@ static int irc_353(struct link_server *server, struct line *line)
 
 	names = irc_line_elem(line, 4);
 
+	int index;
 	while (*names) {
 		long int ovmask = 0;
-		int flagchars = 1;
 		/* some ircds (e.g. unreal) may display several flags for the
                    same nick */
-		while (flagchars) {
-			switch (*names) {
-			case '@':
-				names++;
-				ovmask |= NICKOP;
-				break;
-			case '+':
-				names++;
-				ovmask |= NICKVOICED;
-				break;
-			case '%': /* unrealircd and others: halfop */
-				names++;
-				ovmask |= NICKHALFOP;
-				break;
-			case '&': /* unrealircd: protect */
-			case '~': /* unrealircd: owner */
-				names++;
-				break;
-			default:
-				flagchars = 0;
-				break;
-			}
+		while ((index = bip_get_index(server->prefixes, *names))) {
+			ovmask |= 1 << index;
+			names++;
 		}
 		eon = names;
 		while (*eon && *eon != ' ')
@@ -1454,8 +1444,9 @@ static int irc_367(struct link_server *server, struct line *l)
 }
 
 /* same as irc_315 */
-static int irc_368(struct link_server *server, UNUSED(struct line *l))
+static int irc_368(struct link_server *server, struct line *l)
 {
+	(void)l;
 	struct link *link = LINK(server);
 	if (link->who_client) {
 		if (link->who_client->who_count == 0) {
@@ -1593,6 +1584,7 @@ static int irc_mode(struct link_server *server, struct line *line)
 	int add = 1;
 	unsigned cur_arg = 0;
 	array_t *mode_args = NULL;
+	int ret;
 
 	if (!irc_line_includes(line, 2))
 		return ERR_PROTOCOL;
@@ -1632,99 +1624,76 @@ static int irc_mode(struct link_server *server, struct line *line)
 	 *         ^            ^
 	 *       mode         cur_arg
 	 */
-	const char *nick;
-	long int ovmask;
 	for (mode = irc_line_elem(line, 2); *mode; mode++) {
-		switch (*mode) {
-		case '-':
+		if (*mode == '-')
 			add = 0;
-			break;
-		case '+':
+		else if (*mode == '+')
 			add = 1;
-			break;
-		case 'b':
-			if (!irc_line_includes(line, cur_arg + 3))
-				return ERR_PROTOCOL;
-			cur_arg++;
-			break;
-		case 'o':
-			if (!irc_line_includes(line, cur_arg + 3))
-				return ERR_PROTOCOL;
-			nick = irc_line_elem(line, cur_arg + 3);
+		else {
+			int i = 0;
+			char *str = 0;
 
-			if (!hash_includes(&channel->ovmasks, nick))
-				return ERR_PROTOCOL;
-			ovmask = (long int)hash_remove(&channel->ovmasks, nick);
-			if (add)
-				ovmask |= NICKOP;
-			else
-				ovmask &= ~NICKOP;
-			hash_insert(&channel->ovmasks, nick, (void *)ovmask);
-			cur_arg++;
-			break;
-		case 'h':
-			if (!irc_line_includes(line, cur_arg + 3))
-				return ERR_PROTOCOL;
-			nick = irc_line_elem(line, cur_arg + 3);
+			// Check if mode is known: first user modes then
+			// server modes
+			if (!(str = strchr(server->usermodes, *mode))) {
+				array_each(&server->chanmodes, i, str) {
+					if ((str = strchr(str, *mode)))
+						break;
+				}
+			}
 
-			if (!hash_includes(&channel->ovmasks, nick))
-				return ERR_PROTOCOL;
-
-			ovmask = (long int)hash_remove(&channel->ovmasks, nick);
-			if (add)
-				ovmask |= NICKHALFOP;
-			else
-				ovmask &= ~NICKHALFOP;
-			hash_insert(&channel->ovmasks, nick, (void *)ovmask);
-			cur_arg++;
-			break;
-		case 'v':
-			if (!irc_line_includes(line, cur_arg + 3))
-				return ERR_PROTOCOL;
-			nick = irc_line_elem(line, cur_arg + 3);
-
-			if (!hash_includes(&channel->ovmasks, nick))
-				return ERR_PROTOCOL;
-
-			ovmask = (long int)hash_remove(&channel->ovmasks, nick);
-			if (add)
-				ovmask |= NICKVOICED;
-			else
-				ovmask &= ~NICKVOICED;
-			hash_insert(&channel->ovmasks, nick, (void *)ovmask);
-			cur_arg++;
-			break;
-		case 'k':
-			if (add) {
-				if (!irc_line_includes(line, cur_arg + 3))
-					return ERR_PROTOCOL;
-
-				channel->key = bip_strdup(
-					irc_line_elem(line, cur_arg + 3));
-				cur_arg++;
-			} else {
-				if (channel->key) {
-					free(channel->key);
-					channel->key = NULL;
+			if (str) {
+				// Usermodes, types A & B always take a parameter
+				// Type C take a parameter only when set
+				if (i <= 1 || (i == 2 && add)) {
+					if (!irc_line_includes(line, cur_arg + 3)) {
+						return ERR_PROTOCOL;
+					} else {
+						ret = irc_mode_channel(server, channel, line, mode,
+								add, cur_arg);
+						cur_arg++;
+					}
+				} else {
+					ret = irc_mode_channel(server, channel, line, mode, add,
+							cur_arg);
 				}
 			}
-			break;
-		case 'l':
-			if (add)
-				cur_arg++;
-			break;
-		case 'H':
-		case 'e':
-		case 'q':
-		case 'I':
-			/* Sucky way to try to deal with all the different mode
-			 * out there... */
-			if (irc_line_includes(line, cur_arg + 3))
-				cur_arg++;
-			break;
-		default:
-			break;
 		}
+		if (ret == ERR_PROTOCOL)
+			return ret;
+	}
+	return OK_COPY;
+}
+
+static int irc_mode_channel(struct link_server *s, struct channel *channel,
+				struct line *line, const char* mode, int add, int cur_arg)
+{
+	const char *nick;
+	long int ovmask;
+	int index;
+
+	if (*mode == 'k') {
+		if (add) {
+			channel->key = bip_strdup(
+				irc_line_elem(line, cur_arg + 3));
+		} else {
+			if (channel->key) {
+				free(channel->key);
+				channel->key = NULL;
+			}
+		}
+	} else if ((index = bip_get_index(s->usermodes, *mode))) {
+		nick = irc_line_elem(line, cur_arg + 3);
+
+		if (!hash_includes(&channel->ovmasks, nick))
+			return ERR_PROTOCOL;
+		ovmask = (long int)hash_remove(&channel->ovmasks, nick);
+
+		if (add)
+			ovmask |= 1 << index;
+		else
+			ovmask &= ~(1 << index);
+		hash_insert(&channel->ovmasks, nick, (void *)ovmask);
 	}
 	return OK_COPY;
 }
@@ -2015,7 +1984,7 @@ static struct link_client *irc_accept_new(connection_t *conn)
 
 	ircc = bip_calloc(sizeof(struct link_client), 1);
 	CONN(ircc) = newconn;
-	TYPE(ircc) = IRC_TYPE_LOGING_CLIENT;
+	TYPE(ircc) = IRC_TYPE_LOGGING_CLIENT;
 	CONN(ircc)->user_data = ircc;
 	return ircc;
 }
@@ -2064,7 +2033,10 @@ void irc_client_close(struct link_client *ic)
 			log_client_none_connected(LINK(ic)->log);
 		}
 		irc_client_free(ic);
-	} else if (TYPE(ic) == IRC_TYPE_LOGING_CLIENT) {
+	} else if (TYPE(ic) == IRC_TYPE_TRUST_CLIENT) {
+		unbind_from_link(ic);
+		irc_client_free(ic);
+	} else if (TYPE(ic) == IRC_TYPE_LOGGING_CLIENT) {
 		irc_client_free(ic);
 	}
 }
@@ -2076,7 +2048,7 @@ static void server_setup_reconnect_timer(struct link *link)
 	if (link->last_connection_attempt &&
 			time(NULL) - link->last_connection_attempt
 				< CONN_INTERVAL) {
-		timer = RECONN_TIMER * (link->s_conn_attempt);
+		timer = conf_reconn_timer * (link->s_conn_attempt);
 		if (timer > RECONN_TIMER_MAX)
 			timer = RECONN_TIMER_MAX;
 	}
@@ -2135,9 +2107,29 @@ struct link_server *irc_server_new(struct link *link, connection_t *conn)
 	CONN(s) = conn;
 
 	irc_lag_init(s);
+	array_init(&s->chanmodes);
+	s->prefixes = NULL;
+	s->usermodes = NULL;
+	server_init_modes(s);
+
 	return s;
 }
 
+static void server_init_modes(struct link_server *s)
+{
+	// Default values used if CHANMODES is not specified by the server
+	array_push(&s->chanmodes, bip_strdup("beHIq"));
+	array_push(&s->chanmodes, bip_strdup("k"));
+	array_push(&s->chanmodes, bip_strdup("fjl"));
+	array_push(&s->chanmodes, bip_strdup("fjl"));
+
+	// Default values used if prefix is not specified by the server
+	s->prefixes = bip_realloc(s->prefixes, sizeof(*s->prefixes) * 3);
+	s->usermodes = bip_realloc(s->usermodes, sizeof(s->usermodes) * 3);
+	strcpy(s->prefixes, "@+");
+	strcpy(s->usermodes, "ov");
+}
+
 void irc_server_free(struct link_server *s)
 {
 	if (CONN(s))
@@ -2147,6 +2139,14 @@ void irc_server_free(struct link_server *s)
 	if (s->user_mode)
 		free(s->user_mode);
 
+	int i;
+	char *ptr;
+	array_each(&s->chanmodes, i, ptr)
+		free(ptr);
+
+	MAYFREE(s->prefixes);
+	MAYFREE(s->usermodes);
+
 	hash_iterator_t hi;
 	for (hash_it_init(&s->channels, &hi); hash_it_item(&hi);
 			hash_it_next(&hi)) {
@@ -2173,11 +2173,13 @@ connection_t *irc_server_connect(struct link *link)
 				link->network->serverv[link->cur_server].port,
 				link->vhost, link->bind_port,
 #ifdef HAVE_LIBSSL
-				link->network->ssl, link->ssl_check_mode,
+				link->network->ssl,
+				link->network->ciphers,
+				link->ssl_check_mode,
 				link->user->ssl_check_store,
 				link->user->ssl_client_certfile,
 #else
-				0, 0, NULL, NULL,
+				0, NULL, 0, NULL, NULL,
 #endif
 				CONNECT_TIMEOUT);
 	assert(conn);
@@ -2192,9 +2194,7 @@ connection_t *irc_server_connect(struct link *link)
 	conn->user_data = ls;
 
 	list_add_last(&_bip->conn_list, conn);
-#ifdef HAVE_OIDENTD
 	oidentd_dump(_bip);
-#endif
 	irc_server_startup(ls);
 	return conn;
 }
@@ -2220,6 +2220,14 @@ int irc_server_lag_compute(struct link *l)
 
 void irc_server_shutdown(struct link_server *s)
 {
+	int i;
+	char *cur;
+	array_each(&s->chanmodes, i, cur)
+		free(cur);
+	array_deinit(&s->chanmodes);
+
+	server_init_modes(s);
+
 	if (!s->nick)
 		return;
 	if (LINK(s)->prev_nick)
@@ -2227,15 +2235,16 @@ void irc_server_shutdown(struct link_server *s)
 	LINK(s)->prev_nick = bip_strdup(s->nick);
 }
 
-
-#ifdef HAVE_OIDENTD
-
 #define BIP_OIDENTD_START "## AUTOGENERATED BY BIP. DO NOT EDIT ##\n"
 #define BIP_OIDENTD_END "## END OF AUTOGENERATED STUFF ##\n"
 #define BIP_OIDENTD_END_LENGTH strlen(BIP_OIDENTD_END)
 
 void oidentd_dump(bip_t *bip)
 {
+	mylog(LOG_ERROR, "%d %s", bip->write_oidentd, bip->oidentdpath);
+	if (!bip->write_oidentd || bip->oidentdpath == NULL) {
+		return;
+	}
 	list_iterator_t it;
 	FILE *f;
 	char *bipstart = NULL, *bipend = NULL;
@@ -2342,7 +2351,6 @@ void oidentd_dump(bip_t *bip)
 clean_oidentd:
 	fclose(f);
 }
-#endif
 
 void timeout_clean_who_counts(list_t *conns)
 {
@@ -2473,8 +2481,8 @@ void bip_on_event(bip_t *bip, connection_t *conn)
 
 		line = irc_line_new_from_string(line_s);
 		if (!line) {
-			mylog(LOG_ERROR, "[%s] Error in protocol, closing...",
-					link_name(lc));
+			mylog(LOG_ERROR, "[%s] Can not parse line. Link type: %d. "
+					"closing...", link_name(lc), TYPE(lc));
 			free(line_s);
 			goto prot_err_lines;
 		}
@@ -2483,8 +2491,8 @@ void bip_on_event(bip_t *bip, connection_t *conn)
 		irc_line_free(line);
 		free(line_s);
 		if (r == ERR_PROTOCOL) {
-			mylog(LOG_ERROR, "[%s] Error in protocol, closing...",
-					link_name(lc));
+			mylog(LOG_ERROR, "[%s] Error in protocol. Link type: %d closing...",
+					link_name(lc), TYPE(lc));
 			goto prot_err_lines;
 		}
 		if (r == ERR_AUTH)
@@ -2504,7 +2512,8 @@ prot_err:
 	if (linel)
 		list_free(linel);
 	if (lc) {
-		if (TYPE(lc) == IRC_TYPE_LOGING_CLIENT)
+		if (TYPE(lc) == IRC_TYPE_LOGGING_CLIENT || TYPE(lc) ==
+				IRC_TYPE_TRUST_CLIENT)
 			list_remove(&bip->connecting_client_list, lc);
 		irc_close(lc);
 	}
@@ -2549,10 +2558,8 @@ void irc_main(bip_t *bip)
 		int nc;
 		/* Da main loop */
 		list_t *ready = wait_event(&bip->conn_list, &timeleft, &nc);
-#ifdef HAVE_OIDENTD
 		if (nc)
 			oidentd_dump(bip);
-#endif
 		while ((conn = list_remove_first(ready)))
 			bip_on_event(bip, conn);
 		list_free(ready);
@@ -2623,6 +2630,8 @@ void link_kill(bip_t *bip, struct link *link)
 	while ((ci = list_remove_first(&link->chan_infos_order)))
 		free(ci);
 
+	list_remove(&bip->link_list, link);
+
 	MAYFREE(link->username);
 	MAYFREE(link->realname);
 	MAYFREE(link->s_password);
@@ -2634,3 +2643,89 @@ void link_kill(bip_t *bip, struct link *link)
 	free(link);
 }
 
+static void server_set_chanmodes(struct link_server *l, const char *modes)
+{
+	int i;
+	char *cur;
+	char *dup;
+
+	mylog(LOG_DEBUG, "[%s] Set chanmodes", LINK(l)->name);
+
+	array_each(&l->chanmodes, i, cur)
+		free(cur);
+	array_deinit(&l->chanmodes);
+
+	// handle four categories, ignore all others
+	for (i = 0; i < 4; i++) {
+		cur = strchr(modes, ',');
+		if (cur || modes) {
+			size_t len;
+			if (cur)
+				len = cur - modes;
+			else
+				len = strlen(modes); // last piece
+			dup = bip_malloc(len + 1);
+			memcpy(dup, modes, len);
+			dup[len] = 0;
+			modes = cur + 1;
+		} else {
+			// emptry string
+			dup = bip_calloc(1, sizeof(char));
+		}
+		mylog(LOG_DEBUGVERB, "[%s] Modes: '%s'", LINK(l)->name, dup);
+		array_push(&l->chanmodes, dup);
+	}
+}
+
+
+static void server_set_prefix(struct link_server *s, const char *modes)
+{
+	char * end_mode;
+	unsigned int len;
+
+	mylog(LOG_DEBUG, "[%s] Set user modes", LINK(s)->name);
+
+	// PREFIX=(mode)prefix
+
+	end_mode = strchr(modes + 1, ')'); // skip '('
+	if (*modes != '(' || !end_mode) {
+		mylog(LOG_WARN, "[%s] Unable to parse PREFIX parameter", LINK(s)->name);
+		return;
+	}
+
+	len = end_mode - modes - 1; // len of mode without '('
+	if (len * 2 + 2 != strlen(modes)) {
+		mylog(LOG_WARN, "[%s] Unable to parse PREFIX parameter", LINK(s)->name);
+		return;
+	}
+
+	s->prefixes = bip_realloc(s->prefixes, sizeof(*s->prefixes) * (len + 1));
+	s->usermodes = bip_realloc(s->usermodes, sizeof(s->usermodes) * (len + 1));
+
+	memcpy(s->usermodes, modes + 1, len);
+	s->usermodes[len] = 0;
+	memcpy(s->prefixes, end_mode + 1, len);
+	s->prefixes[len] = 0;
+
+	mylog(LOG_DEBUGVERB, "[%s] user prefix: '%s'", LINK(s)->name, s->prefixes);
+	mylog(LOG_DEBUGVERB, "[%s] user modes: '%s'", LINK(s)->name, s->usermodes);
+}
+
+// Return the position (*1 based*) of car in str, else -1
+static int bip_get_index(const char* str, char car)
+{
+	char *cur;
+	if ((cur = strchr(str, car)))
+		return cur - str + 1;
+	else
+		return 0;
+}
+
+static int bip_fls(int v)
+{
+	unsigned int r = 0;
+	while (v >>= 1)
+		r++;
+
+	return r;
+}
