@@ -36,10 +36,6 @@ struct server {
 
 #define server_new() bip_calloc(sizeof(struct server), 1)
 
-#define NICKOP 1
-#define NICKHALFOP (1<<1)
-#define NICKVOICED (1<<2)
-
 struct channel {
 	char *name;
 	char *mode;
@@ -55,9 +51,15 @@ struct channel {
 
 #define IRC_TYPE_CLIENT (0)
 #define IRC_TYPE_SERVER (1)
-#define IRC_TYPE_LOGING_CLIENT (2)
+#define IRC_TYPE_LOGGING_CLIENT (2)
 #define IRC_TYPE_TRUST_CLIENT (3)
 
+enum BLTimestamp {
+	BLTSNone = 0,
+	BLTSTime,
+	BLTSDateTime,
+};
+
 struct bipuser {
 	/** client connection static data **/
 
@@ -78,10 +80,11 @@ struct bipuser {
 	int backlog_lines;
 	int always_backlog:1;
 	int bl_msg_only:1;
-	int backlog_no_timestamp:1;
 	int blreset_on_talk:1;
 	int blreset_connection:1;
 
+	enum BLTimestamp backlog_timestamp;
+
 #ifdef HAVE_LIBSSL
 	int ssl_check_mode;
 	char *ssl_check_store;
@@ -97,6 +100,7 @@ struct network
 	char *name;
 #ifdef HAVE_LIBSSL
 	int ssl;
+	char *ciphers;
 #endif
 	int serverc;
 	struct server *serverv;
@@ -231,6 +235,13 @@ struct link_server {
 	int lag;
 	int laginit_ts;
 	int lagtest_timeout;
+
+	/* chanmodes */
+	array_t chanmodes;
+
+	/* user modes */
+	char *prefixes;
+	char *usermodes;
 };
 
 typedef struct bip {
@@ -246,10 +257,8 @@ typedef struct bip {
 	hash_t users;
 	list_t errors;
 	struct link_client *reloading_client;
-
-#ifdef HAVE_OIDENTD
 	char *oidentdpath;
-#endif
+	int write_oidentd;
 } bip_t;
 
 void bip_init(bip_t *bip);
