@@ -11,14 +11,13 @@
  * See the file "COPYING" for the exact licensing terms.
  */
 
-#define _XOPEN_SOURCE 500
-
 #include "config.h"
 #include "log.h"
 #include "irc.h"
 #include "util.h"
 #include <sys/time.h>
 #include <stdio.h>
+#include <string.h>
 
 extern int errno;
 extern int log_level;
@@ -84,7 +83,7 @@ int check_dir_r(char *dirname)
 		int slash_ok = 1;
 		while (*tmp == '/') {
 			if (slash_ok) {
-				strncpy(dir + count, "/", 1);
+				strncpy(dir + count, "/", 2);
 				count++;
 				slash_ok = 0;
 			}
@@ -145,10 +144,10 @@ char *log_build_filename(log_t *logdata, const char *destination)
 
 	time(&s);
 	now = localtime(&s);
-	snprintf(year, 5, "%04d", now->tm_year + 1900);
-	snprintf(day, 3, "%02d", now->tm_mday);
-	snprintf(month, 3, "%02d", now->tm_mon + 1);
-	snprintf(hour, 3, "%02d", now->tm_hour);
+	strftime(year, 5, "%Y", now);
+	strftime(day, 3, "%d", now);
+	strftime(month, 3, "%m", now);
+	strftime(hour, 3, "%H", now);
 	snprintf(logfile, MAX_PATH_LEN, "%s/%s", conf_log_root,
 			conf_log_format);
 	replace_var(logfile, "%u", logdata->user->name, MAX_PATH_LEN);
@@ -392,7 +391,8 @@ logstore_t *log_find_file(log_t *logdata, const char *destination)
 
 	time(&t);
 	ltime = localtime(&t);
-	if (ltime->tm_hour != lf->last_log.tm_hour) {
+	if (ltime->tm_year != lf->last_log.tm_year ||
+			ltime->tm_yday != lf->last_log.tm_yday) {
 		logfile_t *oldlf;
 
 		/* day changed, we might want to rotate logfile */
@@ -439,8 +439,8 @@ void log_join(log_t *logdata, const char *ircmask, const char *channel)
 	log_write(logdata, channel, logdata->buffer);
 }
 
-void log_part(log_t *logdata, const const char *ircmask, const char *channel,
-		const const char *message)
+void log_part(log_t *logdata, const char *ircmask, const char *channel,
+		const char *message)
 {
 	if (message)
 		snprintf(logdata->buffer, LOGLINE_MAXLEN,
@@ -712,7 +712,7 @@ void log_reset_all(log_t *logdata)
 		if (ischannel(*hash_it_key(&hi)))
 			log_reset(store);
 		else
-			list_add_last(&drop, strdup(hash_it_key(&hi)));
+			list_add_last(&drop, bip_strdup(hash_it_key(&hi)));
 	}
 
 	char *name;
@@ -853,7 +853,7 @@ char *log_beautify(log_t *logdata, const char *buf, const char *storename,
 	if (!p || !p[0] || !p[1])
 		return _log_wrap(dest, buf);
 	p++;
-	sots = p;
+	sots = logdata->user->backlog_timestamp == BLTSDateTime ? buf : p;
 	p = strchr(p, ' ');
 	if (!p || !p[0] || !p[1])
 		return _log_wrap(dest, buf);
@@ -937,7 +937,7 @@ char *log_beautify(log_t *logdata, const char *buf, const char *storename,
 		strcpy(p, PMSG_ARROW);
 		p += strlen(PMSG_ARROW);
 	}
-	if (logdata->user->backlog_no_timestamp == 0) {
+	if (logdata->user->backlog_timestamp != BLTSNone) {
 		memcpy(p, sots, lots);
 		p += lots;
 		*p++ = '>';
